use std::{
    collections::{HashMap, HashSet, VecDeque},
    pin::pin,
    task::{ready, Poll},
    thread::JoinHandle,
};

use futures::{
    stream::{SplitSink, SplitStream},
    Sink, SinkExt, Stream, StreamExt,
};
use jabber::{client::JabberClient, JID};
use stanza::{
    client::{
        iq::{Iq, IqType, Query},
        Stanza,
    },
    roster,
};
use tokio::sync::mpsc;
use tokio_stream::wrappers::ReceiverStream;
use tokio_util::sync::{PollSendError, PollSender};

pub struct Client {
    client: JabberClient,
    pending_iqs: HashMap<String, mpsc::Sender<Iq>>,
    // database connection (sqlite)
    receiver: ReceiverStream<UpdateMessage>,
    sender: PollSender<CommandMessage>,
}

impl Client {
    pub async fn new(jid: String, password: &str) -> Result<Self, Error> {
        let (read_sender, read_receiver) = mpsc::channel::<UpdateMessage>(20);
        let (write_sender, write_receiver) = mpsc::channel::<CommandMessage>(20);
        let mut jabber_client = JabberClient::new(jid, password)?;
        jabber_client.connect().await?;
        let (write, read) = jabber_client.split();
        let client = Self {
            client: jabber_client,
            receiver: ReceiverStream::new(read_receiver),
            sender: PollSender::new(write_sender),
            pending_iqs: HashMap::new(),
        };
        tokio::spawn(client.process_read(read, read_sender));
        tokio::spawn(client.process_write(write, write_receiver));
        Ok(client)
    }

    pub async fn process_read(
        &self,
        mut stream: SplitStream<JabberClient>,
        sender: mpsc::Sender<UpdateMessage>,
    ) {
        for stanza in stream.next().await {
            tokio::spawn(self.process_stanza(stanza, sender.clone()));
        }
    }

    pub async fn process_write(
        &self,
        mut sink: SplitSink<JabberClient, Stanza>,
        receiver: mpsc::Receiver<CommandMessage>,
    ) {
        for message in receiver.recv_many(, )
    }
}

pub enum Error {
    PollSend(PollSendError<CommandMessage>),
    Jabber(jabber::Error),
}

impl From<jabber::Error> for Error {
    fn from(e: jabber::Error) -> Self {
        Self::Jabber(e)
    }
}

impl Stream for Client {
    type Item = UpdateMessage;

    fn poll_next(
        self: std::pin::Pin<&mut Self>,
        cx: &mut std::task::Context<'_>,
    ) -> std::task::Poll<Option<Self::Item>> {
        pin!(self).receiver.poll_next_unpin(cx)
    }
}

impl Sink<CommandMessage> for Client {
    type Error = Error;

    fn poll_ready(
        self: std::pin::Pin<&mut Self>,
        cx: &mut std::task::Context<'_>,
    ) -> Poll<Result<(), Self::Error>> {
        Poll::Ready(ready!(pin!(self).sender.poll_ready_unpin(cx)))
    }

    fn start_send(self: std::pin::Pin<&mut Self>, item: CommandMessage) -> Result<(), Self::Error> {
        todo!()
    }

    fn poll_flush(
        self: std::pin::Pin<&mut Self>,
        cx: &mut std::task::Context<'_>,
    ) -> Poll<Result<(), Self::Error>> {
        todo!()
    }

    fn poll_close(
        self: std::pin::Pin<&mut Self>,
        cx: &mut std::task::Context<'_>,
    ) -> Poll<Result<(), Self::Error>> {
        todo!()
    }
}

impl From<PollSendError<CommandMessage>> for Error {
    fn from(e: PollSendError<CommandMessage>) -> Self {
        Self::PollSend(e)
    }
}

pub enum CommandMessage {
    Connect,
    GetRoster,
    SendMessage(JID, String),
}

pub enum UpdateMessage {
    Roster(Vec<roster::Item>),
}

impl Client {
    pub async fn process_stanza(
        &mut self,
        stanza: Result<Stanza, jabber::Error>,
        sender: mpsc::Sender<UpdateMessage>,
    ) {
        match stanza {
            Ok(stanza) => todo!(),
            Err(e) => self.process_error(e),
        }
    }

    pub async fn iq(
        &mut self,
        to: Option<JID>,
        r#type: IqType,
        query: Option<Query>,
    ) -> Result<IqResponse, Error> {
        self.client
            .send(Stanza::Iq(Iq {
                from: Some(self.client.jid()),
                // TODO: generate id
                id: "test".to_string(),
                to,
                r#type,
                // TODO: lang
                lang: None,
                query,
                errors: Vec::new(),
            }))
            .await?;
        Ok(todo!())
    }

    pub async fn iq_process(&mut self, iq: Iq) {}
}
